﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Web.Mvc;
using System.Web.Script.Serialization;
using HeuristicLab.Services.Hive.DataAccess;

namespace HeuristicLab.Services.Hive.Statistics.Controllers {
  //[OutputCache(Duration = 60 * 5)]
  public class ChartDataController : Controller {
    private static readonly TimeSpan DefaultDuration = new TimeSpan(1, 0, 0, 0);

    public JsonResult AverageCpuUtilization(DateTime? start = null, DateTime? end = null) {
      using (var db = new HiveDataContext(Settings.Default.HeuristicLab_Hive_LinqConnectionString)) {
        //If no given start date, get date of most recent record
        if (start == null) {
          start = GetRecentDate(db);
        }
        var data =
          from facts in GetClientFacts(db, start, end)
          select new {
            Time = facts.Key,
            CpuUtilization = facts.Average(x => x.CpuUtilization)
          };

        return Json(
          CreateSeriesData(data.ToList(), x => x.Time, x => x.CpuUtilization),
          JsonRequestBehavior.AllowGet);
      }
    }

    public JsonResult UsedCores(DateTime? start = null, DateTime? end = null) {
      using (var db = new HiveDataContext(Settings.Default.HeuristicLab_Hive_LinqConnectionString)) {
        //If no given start date, get date of most recent record
        if (start == null) {
          start = GetRecentDate(db);
        }
        var data =
          from facts in GetClientFacts(db, start, end)
          select new {
            Time = facts.Key,
            UsedCores = facts.Sum(x => x.NumUsedCores),
            TotalCores = facts.Sum(x => x.NumTotalCores)
          };

        return Json(
          CreateSeriesData(data.ToList(), x => x.Time, x => x.UsedCores, x => x.TotalCores),
          JsonRequestBehavior.AllowGet);
      }
    }

    public JsonResult UsedMemory(DateTime? start = null, DateTime? end = null) {
      using (var db = new HiveDataContext(Settings.Default.HeuristicLab_Hive_LinqConnectionString)) {
        //If no given start date, get date of most recent record
        if (start == null) {
          start = GetRecentDate(db);
        }
        var data =
          from facts in GetClientFacts(db, start, end)
          select new {
            Time = facts.Key,
            UsedMemory = facts.Sum(x => x.UsedMemory),
            TotalMemory = facts.Sum(x => x.TotalMemory)
          };

        // Return values in GB
        return Json(
          CreateSeriesData(data.ToList(), x => x.Time, x => x.UsedMemory / 1000, x => x.TotalMemory / 1000),
          JsonRequestBehavior.AllowGet);
      }
    }

    public JsonResult CurrentCores()
    {
      using (var db = new HiveDataContext(Settings.Default.HeuristicLab_Hive_LinqConnectionString))
      {
        var currentCores = (from client in db.FactClientInfos
                            select new { client.NumTotalCores, client.NumUsedCores })
                            .ToList();

        List<int> result = new List<int>();
        result.Add(currentCores.Sum(c => c.NumTotalCores));
        result.Add(currentCores.Sum(s => s.NumUsedCores));

        return Json(result, JsonRequestBehavior.AllowGet);
      }
    }

    public JsonResult CurrentCpuUtilization()
    {
      using (var db = new HiveDataContext(Settings.Default.HeuristicLab_Hive_LinqConnectionString))
      {
        var currentCpuUtil = (from client in db.FactClientInfos
                            select new { client.CpuUtilization })
                            .ToList();

        List<double> result = new List<double>();
        result.Add(Math.Round(currentCpuUtil.Average(s => s.CpuUtilization), 2));

        return Json(result, JsonRequestBehavior.AllowGet);
      }
    }

    public JsonResult CurrentMemory()
    {
      using (var db = new HiveDataContext(Settings.Default.HeuristicLab_Hive_LinqConnectionString))
      {
        var currentMemory = (from client in db.FactClientInfos
                            select new { client.TotalMemory, client.UsedMemory })
                            .ToList();

        List<int> result = new List<int>();
        result.Add(currentMemory.Sum(c => c.TotalMemory)/1000);
        result.Add(currentMemory.Sum(s => s.UsedMemory)/1000);

        return Json(result, JsonRequestBehavior.AllowGet);
      }
    }

    public JsonResult UserTask(string userName, DateTime? start = null, DateTime? end = null, string jobId = null, string taskState=null) {
      using (var db = new HiveDataContext(Settings.Default.HeuristicLab_Hive_LinqConnectionString)) {
        TaskState? state = GetTaskState(taskState);

        var data =
          (from tasks in db.FactTasks
           join jobs in db.DimJobs
             on tasks.JobId equals jobs.JobId
           where jobs.UserName == userName &&
           (!start.HasValue || tasks.StartTime >= start) &&
           (!end.HasValue || tasks.EndTime < end) &&
           (string.IsNullOrEmpty(jobId) || tasks.JobId.ToString() == jobId) &&
           (string.IsNullOrEmpty(taskState) || tasks.TaskState == state)
           select new
           {
             JobName = jobs.JobName,
             JobId = jobs.JobId,
             TaskId = tasks.TaskId,
             TotalWaiting = tasks.TotalWaitingTime,
             TotalTransfer = tasks.TotalTransferTime,
             TotalRuntime = tasks.TotalRuntime,
             StartDate = tasks.StartTime
           }).OrderByDescending(s => s.StartDate).ToList();

        return Json(data, JsonRequestBehavior.AllowGet);
      }
    }

    public JsonResult SlaveInfo(DateTime? start = null, DateTime? end = null, string userName = null, string slaveId=null)
    {
      using (var db = new HiveDataContext(Settings.Default.HeuristicLab_Hive_LinqConnectionString))
      {
        var data =
          (from slaves in db.FactClientInfos
           join users in db.DimUsers
            on slaves.UserId equals users.UserId
           join clients in db.DimClients
            on slaves.ClientId equals clients.Id
           where (!start.HasValue || slaves.Time >= start) &&
           (!end.HasValue || slaves.Time < end) &&
           (string.IsNullOrEmpty(userName) || users.Name == userName) &&
           (string.IsNullOrEmpty(slaveId) || slaves.ClientId.ToString() == slaveId)
           select new
           {
             ClientName = clients.Name,
             SlaveID = slaves.ClientId,
             Time = slaves.Time,
             UsedCores = slaves.NumUsedCores,
             TotalCores = slaves.NumTotalCores,
             UsedMemory = slaves.UsedMemory,
             TotalMemory = slaves.TotalMemory,
             CPUUtilization = slaves.CpuUtilization
           }).OrderByDescending(s => s.Time).GroupBy(s => s.SlaveID).ToList();

        return Json(data, JsonRequestBehavior.AllowGet);
      }
    }

    private static TaskState? GetTaskState(string taskState) {
      TaskState? state;
      switch (taskState)
      {
        case "Aborted":
          state = TaskState.Aborted;
          break;
        case "Calculating":
          state = TaskState.Calculating;
          break;
        case "Failed":
          state = TaskState.Failed;
          break;
        case "Finished":
          state = TaskState.Finished;
          break;
        case "Offline":
          state = TaskState.Offline;
          break;
        case "Paused":
          state = TaskState.Paused;
          break;
        case "Transferring":
          state = TaskState.Transferring;
          break;
        case "Waiting":
          state = TaskState.Waiting;
          break;
        default :
          state = null;
          break;
      }
      return state;
    }

    private static IOrderedQueryable<IGrouping<DateTime, FactClientInfo>> GetClientFacts(HiveDataContext db, DateTime? start = null, DateTime? end = null) {
      start = start ?? DateTime.Now - DefaultDuration;
      end = end ?? DateTime.Now;

      return from ci in db.FactClientInfos
             where ci.Time >= start && ci.Time < end
             group ci by ci.Time into timeGroup
             orderby timeGroup.Key
             select timeGroup;
    }

    private static IEnumerable<IEnumerable<object[]>> CreateSeriesData<T>(IEnumerable<T> data, Func<T, DateTime> timeSelector, params Func<T, object>[] seriesSelectors) {
      return seriesSelectors.Select(selector =>
        data.Select(x => new[] {
          timeSelector(x).ToUniversalTime().ToUnixTimestamp(),
          selector(x)
        })
      );
    }

    private static DateTime GetRecentDate(HiveDataContext db)
    {
      var mostRecent = from dates in db.FactClientInfos
        group dates by dates.Time into startDate
        select startDate.OrderByDescending(t=>t.Time).FirstOrDefault();
      return mostRecent.Max(t=>t.Time);
    }
  }

  public static class DateTimeExtensions {
    public static long ToUnixTimestamp(this DateTime dateTime) {
      var duration = dateTime - new DateTime(1970, 1, 1, 0, 0, 0);

      return (long)duration.TotalMilliseconds;
    }
  }
}
